#ifndef _PLUMP_H
#define _PLUMP_H

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Error codes
 */
#define PLUMP_SUCCESS                       0
#define PLUMP_SUCCESS_SRCBREAKPOINTHIT      1
#define PLUMP_SUCCESS_USRBREAKPOINTHIT      2

#define PLUMP_ERROR_FAILURE                 (-1)
#define PLUMP_ERROR_NOTSUPPORTED            (-2)
#define PLUMP_ERROR_NOACCESS                (-3)
#define PLUMP_ERROR_NOSUCHPID               (-4)
#define PLUMP_ERROR_NOSUCHADDR              (-5)
#define PLUMP_ERROR_NOSUCHREGISTER          (-6)
#define PLUMP_ERROR_NOMEMORY                (-7)
#define PLUMP_ERROR_BADPTR                  (-8)
#define PLUMP_ERROR_MOREDATA                (-9)
#define PLUMP_ERROR_NOTSTOPPED              (-10)
#define PLUMP_ERROR_NOTSUPPORTEDMEDIUM      (-11)
#define PLUMP_ERROR_ALREADY                 (-12)

/*
 * Plump access rights
 */
#define PLUMP_ACCESS_READ                   (1 << 0)
#define PLUMP_ACCESS_WRITE                  (1 << 1)
#define PLUMP_ACCESS_EXEC                   (1 << 2)
#define PLUMP_ACCESS_ALL                    0xF

/*
 * Plump state
 */
#define PLUMP_STATE_STOPPED                 (1 << 0)

/*
 * Image types
 */
#define PLUMP_IMAGE_TYPE_ELF                0x0001
#define PLUMP_IMAGE_TYPE_AOUT               0x0002
#define PLUMP_IMAGE_TYPE_PE                 0x0003

/*
 * Continue flags
 */
#define PLUMP_CONTINUE_FLAG_INDEFINITELY    (1 << 0) // Run indefinitely
#define PLUMP_CONTINUE_FLAG_SINGLESTEP      (1 << 1) // Singlestep
#define PLUMP_CONTINUE_FLAG_UNTILSRCBP      (1 << 2) // Until a source breakpoint is hit
#define PLUMP_CONTINUE_FLAG_UNTILUSRBP      (1 << 3) // Until a user supplied breakpoint is hit

/*
 * Register bits
 */
#define REGISTER_BITS(name)                 (name &  0xF0000000)
#define REGISTER_BITS_8                     (0x1  << 28)
#define REGISTER_BITS_16                    (0x2  << 28)
#define REGISTER_BITS_32                    (0x3  << 28)
#define REGISTER_BITS_64                    (0x4  << 28)
/*
 * Register types
 */
#define REGISTER_TYPE(name)                 (name  &  0x0FFF0000)
#define REGISTER_TYPE_STACK_POINTER         (0x001 << 16)
#define REGISTER_TYPE_FRAME_POINTER         (0x002 << 16)
#define REGISTER_TYPE_PROGRAM_COUNTER       (0x003 << 16)
#define REGISTER_TYPE_UNSPECIFIED           (0xFFF << 16)

/*
 * Register names
 *
 * 0x1FFF0001
 *       ^^^^--- specific name
 *    ^^^------- type
 *   ^---------- bits
 */
#define REGISTER_NAME(name)                 (name & 0xFFFF)

// IA32
#define REGISTER_IA32_EAX                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0001)
#define REGISTER_IA32_EBX                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0002)
#define REGISTER_IA32_ECX                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0003)
#define REGISTER_IA32_EDX                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0004)
#define REGISTER_IA32_EDI                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0005)
#define REGISTER_IA32_ESI                   (REGISTER_BITS_32 | REGISTER_TYPE_UNSPECIFIED     | 0x0006)
#define REGISTER_IA32_EBP                   (REGISTER_BITS_32 | REGISTER_TYPE_FRAME_POINTER   | 0x0007)
#define REGISTER_IA32_ESP                   (REGISTER_BITS_32 | REGISTER_TYPE_STACK_POINTER   | 0x0008)
#define REGISTER_IA32_EIP                   (REGISTER_BITS_32 | REGISTER_TYPE_PROGRAM_COUNTER | 0x0009)

#define REGISTER_IA32_SS                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0020)
#define REGISTER_IA32_CS                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0021)
#define REGISTER_IA32_DS                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0022)
#define REGISTER_IA32_ES                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0023)
#define REGISTER_IA32_FS                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0024)
#define REGISTER_IA32_GS                    (REGISTER_BITS_16 | REGISTER_TYPE_UNSPECIFIED     | 0x0025)

typedef unsigned char      pubyte_t;      // 8 bit  unsigned char
typedef unsigned short     puint16_t;     // 16 bit unsigned short
typedef unsigned long      puint32_t;     // 32 bit unsigned integer

typedef unsigned long      plumpaddr_t;   // Virtual address
typedef unsigned long      plumpimage_t;  // Image type

struct _plump;

typedef enum _plumpmedium {
	PLUMP_MEDIUM_ANY    = 0, // Select at runtime
	PLUMP_MEDIUM_PTRACE = 1, // ptrace            (OS: *nix)
	PLUMP_MEDIUM_PROCFS = 2, // Proc FS           (OS: *nix)
	PLUMP_MEDIUM_DBGAPI = 3, // Debug API         (OS: Windows)
} plumpmedium_t;

typedef struct _plumpreg {

	int name;

	union
	{
		pubyte_t  b8;
		puint16_t b16;
		puint32_t b32;
	} val;

} plumpreg_t;

typedef struct _plump_functions {

	plumpmedium_t (*getMedium)();

	int (*open)(struct _plump *p, int pid, unsigned long access);
	int (*close)(struct _plump *p);

	int (*execContinue)(struct _plump *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags);
	int (*execStop)(struct _plump *p);

	int (*memoryWrite)(struct _plump *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesWritten);
	int (*memoryRead)(struct _plump *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesRead);
	int (*memoryIsValid)(struct _plump *p, plumpaddr_t addr);
	int (*memoryGetEndOfSegment)(struct _plump *p, plumpaddr_t addr, plumpaddr_t *end);

	int (*registersSet)(struct _plump *p, plumpreg_t *registers, unsigned long numRegisters);
	int (*registersGet)(struct _plump *p, plumpreg_t *registers, unsigned long numRegisters);

} plumpfunc_t;

typedef struct _plumpbp {

	plumpaddr_t   addr;
	unsigned long enabled;

} plumpbp_t;

typedef struct _plump {

	/**
	 * Medium virtual function table
	 */
	plumpfunc_t   vftable;
	/**
	 * Medium in use
	 */
	plumpmedium_t medium;

	/**
	 * Process Id being operated on.
	 */
	int           pid;
	/**
	 * The current state of the object. (OPENED, STOPPED, etc)
	 */
	unsigned long state;
	/**
	 * Access rights
	 */
	unsigned long access;

	/**
	 * Breakpoint array
	 */
	plumpbp_t     *userBreakpoints;
	/**
	 * Number of user breakpoints
	 */
	unsigned long numUserBreakpoints;

	/**
	 * Data space for mediums to hold their data
	 */
	void          *mediumExtension;

} plump_t;

int plump_open(int pid, unsigned long access, plumpmedium_t medium, plump_t **ret);
int plump_getMedium(plump_t *p, plumpmedium_t *medium);
int plump_close(plump_t *p);

int plump_execContinue(plump_t *p, plumpaddr_t start, plumpaddr_t *stop, unsigned long flags);
int plump_execStop(plump_t *p);

int plump_breakpointAdd(plump_t *p, plumpaddr_t addr);
int plump_breakpointIsWatching(plump_t *p, plumpaddr_t addr);
int plump_breakpointEnable(plump_t *p, plumpaddr_t addr);
int plump_breakpointDisable(plump_t *p, plumpaddr_t addr);
int plump_breakpointRemove(plump_t *p, plumpaddr_t addr);

int plump_imageGetType(plump_t *p, plumpimage_t *image);
int plump_imageGetPath(plump_t *p, char *path, unsigned long *pathSize);

int plump_memoryRead(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesRead);
int plump_memoryWrite(plump_t *p, plumpaddr_t addr, void *buffer, unsigned long bufferLength, unsigned long *bytesWritten);
int plump_memoryIsValid(plump_t *p, plumpaddr_t addr);
int plump_memoryGetEndOfSegment(plump_t *p, plumpaddr_t addr, plumpaddr_t *end);

int plump_registersSet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters);
int plump_registersGet(plump_t *p, plumpreg_t *registers, unsigned long numRegisters);
int plump_registersGetStackPointer(plump_t *p, plumpreg_t *sp);
int plump_registersSetStackPointer(plump_t *p, plumpreg_t *sp);
int plump_registersGetFramePointer(plump_t *p, plumpreg_t *fp);
int plump_registersSetFramePointer(plump_t *p, plumpreg_t *fp);
int plump_registersGetProgramCounter(plump_t *p, plumpreg_t *pc);
int plump_registersSetProgramCounter(plump_t *p, plumpreg_t *pc);

#ifdef __cplusplus
}
#endif

#endif
