/*
    dltrace 0.5
    Copyright (C) 2005 Richard Johnson <rjohnson@idefense.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <stdarg.h>
#include "dltrace.h"
#include "plump.h"

void usage(char *argv)
{
	printf(
	"Usage: %s [OPTIONS] -- [file path] [args ...]\n"
	"  -p pid       attach to process id (file path required)\n"
	"\nTrace Options\n"
	"  -l           trace library calls\n"
//	"  -s           trace system calls\n"
//	"  -b           trace logical blocks\n"
	"  -d depth     limit trace recursion depth\n"
	"\nOutput Options\n"
	"  -r           display registers (use multiple times)\n"
	"  -t           tree mode \n"
	"  -v           verbose (use multiple times)\n"
	"  -o file      write trace output to file    \n"
//	"  -x file      create xml output file\n"
	"\n", argv);
	exit(0);
}

// ^C (SIGINT) handler for detaching cleanly
void sigint_h(int sig)
{
	list_t  *p;
	proc_t  *ps;
	
	vlog(stderr, "\ndltrace: caught sigint!\n");
	if(!options.attach)
	{
		vlog(stdout, "dltrace: exiting..\n");
		exit(0);
	}	
	p = pslist;
	do
	{
		ps = p->data;
		if(plump_close(ps->plump) == PLUMP_SUCCESS)
			vlog(stdout, "dltrace: [%d] %s detached successfully.\n", ps->pid, ps->argv[0]);
		else 
			vlog(stdout, "dltrace: [%d] %s detach was unsuccessful. Program in unstable state.\n", ps->pid, ps->argv[0]);
		p = p->next;
	} 
	while(p->next != NULL);
	
	exit(0);
}
		

int main(int argc, char **argv)
{
	uint32_t c;
	proc_t *ps;
	plumpreg_t *regs;
	
	if(argc < 2)
		usage(argv[0]);

	options.depth = MAX_CALLDEPTH;
	ps = malloc(sizeof(proc_t));
	if(!ps)
	{
		perror("malloc");
		return -1;
	}
	memset(ps, 0, sizeof(proc_t));

	while((c = getopt (argc, argv, "p:lsbd:rtvo:x:")) != -1)
	{
		switch(c)
		{
			case 'p': // trace ps->pid 
				options.attach++;
				ps->pid = atoi(optarg);
				ps->argv++;
				break;
			case 'l': // print library calls
				options.ltrace++;
				break;
			case 's': // print syscalls
				options.strace++;
				break;
			case 'b': // block level tracing
				options.btrace++;
				break;
			case 'd':
				options.depth = (unsigned int)atoi(optarg);
				ps->argv++;
				if(options.depth < 1)
					usage(argv[0]);
				if(options.verbose > 1)
					vlog(stdout, "Call recurision depth limited to %d calls.\n", options.depth);
				break;
			case 'r': // show registers
				options.regs++;
				break;
			case 't': // show registers
				options.tree++;
				break;
			case 'v': // verbosity
				options.verbose++;
				break;
			case 'o':
				options.logfile = fopen(optarg, "w+");
				ps->argv++;
				if(!options.logfile)
				{
					perror("fopen");
					return -1;
				}
				break;
#if 0				
			case 'x':
				options.xmlfile = fopen(optarg, "w+");
				ps->argv++;
				if(!options.xmlfile)
				{
					perror("fopen");
					return -1;
				}
				break;
#endif				
			default:
				usage(argv[0]);
		}
	}
	ps->argv = argv + optind;
	
	if(!(strncmp(argv[optind - 1], "--", 2) == 0))
	{
		vlog(stderr, "error: path for traced binary is required\n");
		usage(argv[0]);
	}

	if(!options.ltrace && !(options.btrace || options.strace))
		options.ltrace++;
	
	regs = ps->regs;
	regs[0].name = REGISTER_IA32_EAX;
	regs[1].name = REGISTER_IA32_EBX;
	regs[2].name = REGISTER_IA32_ECX;
	regs[3].name = REGISTER_IA32_EDX;
	regs[4].name = REGISTER_IA32_EDI;
	regs[5].name = REGISTER_IA32_ESI;
	regs[6].name = REGISTER_IA32_ESP;
	regs[7].name = REGISTER_IA32_EBP;
	regs[8].name = REGISTER_IA32_EIP;
	regs[9].name  = REGISTER_IA32_SS;
	regs[10].name = REGISTER_IA32_CS;
	regs[11].name = REGISTER_IA32_DS;
	regs[12].name = REGISTER_IA32_ES;
	regs[13].name = REGISTER_IA32_FS;
	regs[14].name = REGISTER_IA32_GS;

	if(ps->pid == 0)
	{
		switch(ps->pid = fork())
		{
			case -1:
				perror("fork");
				return -1;
			case  0: // child 
			{
				kill(getpid(), SIGSTOP);
				execv(ps->argv[0], (char **)ps->argv);
				break;
			}
			default:
				kill(ps->pid, SIGSTOP);
				break;
		}
	}

	vlog(stderr, "dltrace: attaching to process [%d]\n", ps->pid);
	if(dt_open_process(ps) < 0)
	{
		vlog(stderr, "error: failed to open process [%d]\n", ps->pid);
		exit(-1);
	}

	if(dt_init_trace(ps) < 0)
	{
		vlog(stderr, "error: failed to init trace on process [%d]\n", ps->pid);
		exit(-1);
	}

	pslist = malloc(sizeof(list_t));
	if(!pslist)
	{
		perror("malloc");
		exit(-1);
	}
	memset(pslist, 0, sizeof(list_t));
	list_push(&pslist, ps);

	// register signal handler so we can replace breakpoints on detach
	if (signal(SIGINT, sigint_h) == SIG_ERR) {
		perror("signal");
		exit(-1);
	}

	return dt_run_trace(ps);	
}




// handle multiple output streams
void vlog(FILE *fd, const char *fmt, ...)
{
	va_list args; 
	
	va_start(args, fmt);
	vfprintf(fd, fmt, args);
	if(options.logfile && fd != stderr)
		vfprintf(options.logfile, fmt, args);
	va_end(args);
}

